/**
	AOK Trigger Studio (See aokts.cpp for legal conditions.)
	WINDOWS VERSION
	scen.cpp -- Scenario wrapper code
**/

#include "scen.h"
#include "zlibfile.h"
#include "utilio.h"
#include "settings.h"
#include <direct.h>
#include <string.h>
#include <math.h>

/* some options */
#define BMP_READ_HACK

class Scenario scen;

#define CHECKEOF() \
	if (feof(dcin)) \
	{ \
		ret = ERR_data; \
		strcpy(msg, e_data); \
		printf("Scenario Error: early end-of-file.\n"); \
		goto Cleanup; \
	}

/* PerVersions */
const Scenario::PerVersion Scenario::pv1_18 =
{
	5,
	true,
	30, 20
};

const Scenario::PerVersion Scenario::pv1_22 =
{
	6,
	true,
	30, 20
};

const Scenario::PerVersion Scenario::pv1_30 =
{
	6,
	true,
	60, 60
};

/* The Scenario */

#define SMSG	" For stability\'s sake, I must discontinue reading the scenario." \
				"\nPlease contact cyan.spam@gmail.com regarding this error."

const char e_none[]	= "Scenario operation completed successfully.";
const char e_data[]	= "Sorry, I've encountered a problem with the scenario data." SMSG;
const char e_compr[]	= "Zlib compression/decompression failed.";
const char e_mem[]	= "I ran out of memory. This probably means "\
	"that the scenario is invalid or that I got out of sync.";
const char e_zver[]	= "Your version of zlib1.dll is incompatible. I need at least %s. ";
const char e_digit[]	= "Oops, my programmer (David Tombs) made an error.\nProblem: %s" SMSG;
const char e_bitmap[]	= "Sorry, the bitmap data in this scenario appears incomplete." SMSG;
const char e_sver[]	= "%s is of an unrecognized format version." SMSG;

/* Game-specific settings */
#if (GAME == 1)
const char e_noscen[]	= "%s isn't an Age of Empires scenario. Loading terminated.";
#define PLAYER_POP_UNREAD_COUNT 6

#elif (GAME == 2)
const char e_noscen[]	= "%s isn't a Star Wars: Galactic Battlegrounds scenario. Loading terminated.";
#define PLAYER_POP_UNREAD_COUNT 5

#endif

Scenario::Scenario()
:	mod_status(false),
	next_uid(0), bBitmap(false), files(NULL)
{
//	strcpy(mark, "DGT");
	strcpy(players[GAIA_INDEX].name, "GAIA");
}

Scenario::~Scenario()
{
	delete [] files;
}

void Scenario::reset()
{
	int i;
	
	/* Internal */
	mod_status = false;
	ver = VER_AOE2TC;

	/* "Blank" scenario */
	strcpy(header.version, "1.21");
	next_uid = 0;
	ver2 = (float)1.22;
	memset(origname, 0, sizeof(origname));

	for (i = 0; i < 6; i++)
	{
		messages[i].erase();
		mstrings[i] = -1;
	}

	memset(cinem, 0, sizeof(cinem));
	bBitmap = false;

	for (i = 0; i < NUM_PLAYERS; i++)
		players[i].reset();
	players[0].enable = true;
	players[0].human = true;
	Player::num_players = 9;

	memset(&vict, 0, sizeof(vict));
	map.reset();

	triggers.clear();
	t_order.clear();

	files = NULL;

	memset(msg, 0, sizeof(msg));
}

Scenario::AOKBMP::AOKBMP()
:	colors(NULL), image(NULL)
{
	//prep fileheader
	file_hdr.bfType = 0x4D42;	//"MB"
	file_hdr.bfReserved1 = 0;
	file_hdr.bfReserved2 = 0;

	/* these are set later when we read the BMP from the scenario */
	file_hdr.bfSize = 0;
	file_hdr.bfOffBits = 0;
}

Scenario::AOKBMP::~AOKBMP()
{
	delete [] colors;
	delete [] image;
}

bool Scenario::AOKBMP::read(FILE *in)
{
	bool ret = true;

	ret = (fread(&info_hdr, sizeof(BITMAPINFOHEADER), 1, in) == 1);

	/* some ugly hacks because Genie has horrible BMP support */
#ifdef BMP_READ_HACK
	info_hdr.biClrUsed = 256;
	info_hdr.biClrImportant = 256;
	info_hdr.biSizeImage = info_hdr.biWidth * info_hdr.biHeight;
#endif

	file_hdr.bfOffBits = (sizeof(BITMAPFILEHEADER) + info_hdr.biSize + info_hdr.biClrUsed * sizeof(RGBQUAD));
	file_hdr.bfSize = (file_hdr.bfOffBits + info_hdr.biSizeImage);

	colors = new RGBQUAD[info_hdr.biClrUsed];
	fread(colors, sizeof(RGBQUAD), info_hdr.biClrUsed, in);

	image = new char[info_hdr.biSizeImage];
	fread(image, sizeof(char), info_hdr.biSizeImage, in);

	return ret;
}

void Scenario::AOKBMP::write(FILE *out)
{
	fwrite(&info_hdr, sizeof(info_hdr), 1, out);
	fwrite(colors, sizeof(RGBQUAD), info_hdr.biClrUsed, out);
	fwrite(image, sizeof(char), info_hdr.biSizeImage, out);
}

bool Scenario::AOKBMP::toFile(const char *path)
{
	FILE *bmp_out;

	if (!file_hdr.bfSize)
		return false;

	bmp_out = fopen(path, "wb");
	if (!bmp_out)
		return false;

	fwrite(&file_hdr, sizeof(BITMAPFILEHEADER), 1, bmp_out);
	fwrite(&info_hdr, sizeof(BITMAPINFOHEADER), 1, bmp_out);
	fwrite(colors, sizeof(RGBQUAD), info_hdr.biClrUsed, bmp_out);
	fwrite(image, info_hdr.biSizeImage, 1, bmp_out);

	fclose(bmp_out);
	return true;
}

void Scenario::AOKBMP::reset()
{
	file_hdr.bfSize = 0;
	delete [] colors;
	delete [] image;
}

bool Scenario::isExpansion()
{
	bool ret;

#if (GAME == 1)
	switch (scen.ver)
	{
	case VER_AOE1:
	case VER_AOE2:
	default:
		ret = false;
		break;

	case VER_AOE2TC:
		ret = true;
	}
#elif (GAME == 2)
	switch (scen.ver)
	{
	case VER_AOE2TC:
	default:
		ret = false;
	}
#endif

	return ret;
}

char Scenario::StandardAI[] = "RandomGame";

/* Open a scenario, read header, and read compressed data */

int Scenario::open(const char *path, const char *dpath)
{
	FILE *scx;			//SCX file
	FILE *tempout;		//decompressed output file
	unsigned char *compressed = NULL;	//buffer for compressed data
	int clen;	//length of compressed data
	int code;	//error codes

	/* Initialization */

	mod_status = false;
	memset(msg, 0, sizeof(msg));
	if (!(scx = fopen(path, "rb")))
	{
		printf("Failed opening %s.", path);
		sprintf(msg, e_noscen, path);
		return ERR_noscen;
	}

	/* Header */
	if (!header.read(scx))
	{
		printf("Invalid scenario header in %s", path);
		sprintf(msg, e_noscen, path);
		code = ERR_noscen;
		goto Cleanup;
	}

	switch (*(short*)(header.version + 2))
	{
#ifdef _DEBUG	//testing AOE1 support in debug mode
	case 0x3031:	//10
		ver = VER_AOE1;
		break;

	case 0x3131:	//11
		ver = VER_AOE1;
		break;
#endif

	case 0x3831:	//18
		ver = VER_AOE2;
		break;

	case 0x3132:	//21
		ver = VER_AOE2TC;
		break;

	default:
		printf("Unrecognized scenario version: %s.", header.version);
		code = ERR_sver;
		sprintf(msg, e_sver, path);
		goto Cleanup;
	}

	/* Inflate and Read Compressed Data */
	clen = fsize(path) - (header.length + 8);	//subtract header length

	if (clen > 0)
		compressed = new unsigned char[clen];

	if (compressed)
	{
		tempout = fopen(dpath, "wb");

		fread(compressed, sizeof(char), clen, scx);
		
		code = inflate_file(compressed, clen, tempout);

		fclose(tempout);	//must be closed before data is read

		//translate from zlib codes to AOKTS_ERROR codes
		switch (code)
		{
		case Z_OK:
		case Z_STREAM_END:
			code = ERR_none;
			break;

		case Z_STREAM_ERROR:
			code = ERR_digit;
			sprintf(msg, e_digit, "zlib stream error");
			break;

		case Z_DATA_ERROR:
			code = ERR_compr;
			printf("Decompression of %s failed. Corrupted scenario?.\n", path);
			strcpy(msg, e_compr);
			break;

		case Z_MEM_ERROR:
			code = ERR_mem;
			printf("Not enough memory for decompression of %s.\n", path);
			strcpy(msg, e_mem);
			break;

		case Z_BUF_ERROR:
			code = ERR_digit;
			sprintf(msg, e_digit, "zlib buffer error.");
			break;

		case Z_VERSION_ERROR:
			code = ERR_zver;
			sprintf(msg, e_zver, ZLIB_VERSION);
			break;
		}

		if (code == ERR_none)
			code = read_data(dpath);
	}
	else
	{
		code = ERR_mem;
		printf("Could not allocate %d bytes of memory to read compressed file.\n",
			clen);
		strcpy(msg, e_mem);
	}

	/* Cleanup & Return */
Cleanup:

	delete [] compressed;
	fclose(scx);
	fflush(stdout);	//report errors to logfile

	return code;
}

/* Open a destination scenario, write the header, and optionally write compressed data */

int Scenario::save(const char *path, const char *dpath, bool write)
{
	unsigned char *uncompressed;
	size_t uc_len;
	FILE *scx;
	FILE *temp;
	int code = 0;	//return from zlib functions

	scx = fopen(path, "wb");

	/* Header */
	header.write(scx, messages, getPlayerCount());

	/* Write uncompressed data to a temp file */
	if (write)
		write_data(dpath);

	/* Then compress and write it to the scenario */
	uc_len = fsize(dpath);
	uncompressed = new unsigned char[uc_len];
	temp = fopen(dpath, "rb");
	fread(uncompressed, sizeof(unsigned char), uc_len, temp);
	fclose(temp);

	code = deflate_file(uncompressed, uc_len, scx);
	
	/* translate zlib codes to AOKTS_ERROR codes */
	switch (code)
	{
	case Z_OK:
	case Z_STREAM_END:
		code = ERR_none;
		break;

	case Z_STREAM_ERROR:
		code = ERR_digit;
		sprintf(msg, e_digit, "zlib stream error");
		break;

	case Z_MEM_ERROR:
		code = ERR_mem;
		printf("Could not allocate memory for compression stream.\n");
		strcpy(msg, e_mem);
		break;

	case Z_BUF_ERROR:
		code = ERR_digit;
		sprintf(msg, e_digit, "zlib buffer error.");
		break;

	case Z_VERSION_ERROR:
		code = ERR_zver;
		sprintf(msg, e_zver, ZLIB_VERSION);
		break;
	}

	delete [] uncompressed;
	fclose(scx);
	fflush(stdout);	//report errors to logfile

	return code;
}

#define UNREAD(f, s) fseek(f, s, SEEK_CUR)

bool Scenario::_header::read(FILE *scx)
{
	bool ret = true;
	long check;
	unsigned long len;	//Instructions string length

	fread(version, sizeof(char), 4, scx);

	fread(&length, sizeof(long), 1, scx);

	fread(&check, sizeof(long), 1, scx);
	REPORTS(check == 2, ret = false, "Header check value invalid.");

	fread(&timestamp, sizeof(time_t), 1, scx);

	fread(&len, sizeof(long), 1, scx);
	UNREAD(scx, len);	//instructions

	UNREAD(scx, sizeof(long) * 2);	//unknown & playercount

	return ret;
}

void Scenario::_header::write(FILE *scx, const SString *instr, long players)
{
	long num;

	fwrite(version, sizeof(char), 4, scx);

	/* Length calculation is a little tricky */
	length = 0x14 + instr->lwn();
	fwrite(&length, sizeof(long), 1, scx);

	num = 2;
	fwrite(&num, sizeof(long), 1, scx);

	fwrite(&timestamp, sizeof(time_t), 1, scx);

	instr->write(scx, sizeof(long));

	NULLS(scx, sizeof(long));

	fwrite(&players, sizeof(long), 1, scx);
}

void Scenario::_header::reset()
{
	memset(version, 0, sizeof(version));
	length = 0;
	time(&timestamp);
}

/* Check the "good" variable, report if false. */
#define CHECK(string) \
	if (!good) \
	{ \
		ret = ERR_data; \
		strcpy(msg, e_data); \
		printf(string); \
		goto Cleanup; \
	}
#define CHECKA(string, arg) \
	if (!good) \
	{ \
		ret = ERR_data; \
		strcpy(msg, e_data); \
		printf(string, arg); \
		goto Cleanup; \
	}

/*
	FEP = For Each Player.

	Requires an int i. Takes p from player 1 through player 16.
*/
#define FEP(p) \
	for (i = PLAYER1_INDEX, p = players; i < NUM_PLAYERS; i++, p++)

#define UNKNOWN(type, expected, name) \
	{ \
		type n; \
		fread(&n, sizeof(type), 1, dcin); \
		if (n != expected) \
			printf("Unusual " name " value: %d.\n", n); \
	}

inline int myround(float n)
{
	return int((n >= 0.0) ? n + 0.5 : n - 0.5);
}

int Scenario::read_data(const char *path)	//decompressed data
{
	AOKTS_ERROR ret = ERR_none;
	int i;
	Player *p;
	float f;
	FILE *dcin;	//temporary decompressed output file
	bool good = true;

	printf("Read decompressed data file %s...\n", path);
	dcin = fopen(path, "rb");

	/* Compressed Header */

	fread(&next_uid, sizeof(long), 1, dcin);
	fread(&ver2, sizeof(float), 1, dcin);

	/* Set PerVersion according to ver2 */
	switch (myround(ver2 * 100))
	{
	case 118:
	case 119:
	case 120:
	case 121:
		perversion = &pv1_18;
		break;

	case 122:
		perversion = &pv1_22;
		break;

	case 130:
		perversion = &pv1_30;
		break;

	default:
		ret = ERR_sver;
		sprintf(msg, e_sver, path);
		goto Cleanup;
	}

	//player names
	FEP(p)
	{
		fread(p->name, sizeof(char), sizeof(p->name), dcin);
		fseek(dcin, 0x100 - sizeof(p->name), SEEK_CUR);	//jump to the end of the 256-byte name area
	}

	//string tables
	if (ver >= VER_AOE2)
	FEP(p)
		fread(&p->stable, 1, sizeof(long), dcin);

	//player data 1
	FEP(p)
	{
		fread(&p->enable, sizeof(long), 3, dcin);	//enable, human, civ
		UNKNOWN(long, 4, "PlayerData1 uknown");
	}

	fread(&unknown, sizeof(long), 1, dcin);
	UNKNOWN(char, 0, "post-PlayerData1 char");

	fread(&f, sizeof(float), 1, dcin);
	if (f != -1)
	{
		ret = ERR_data;
		strcpy(msg, e_data);
		printf("Unrecognized unknown3 value. Probably reading out of sync.\n");
	}

	READCS(dcin, origname, sizeof(short));

	/* Messages & Cinematics */
	if (perversion->mstrings)
		fread(mstrings, sizeof(long), perversion->messages_count, dcin);

	for (i = 0; i < perversion->messages_count; i++)
			messages[i].read(dcin, sizeof(short));

	for (i = 0; i < NUM_CINEM; i++)
		READCS(dcin, cinem[i], sizeof(short));

	//bitmap
	fread(&bBitmap, sizeof(long), 1, dcin);
	UNREAD(dcin, sizeof(long) * 2);	//x, y (duplicate)
	UNKNOWN(short, -1, "bitmap unknown");

	if (bBitmap && !bitmap.read(dcin))	//note: won't check the 2nd condition if 1st fails
	{
		ret = ERR_bitmap;
		strcpy(msg, e_bitmap);
	}

	/* Player Data 2 */

	for (i = 0; i < NUM_UNK; i++)
		unk[i].read(dcin, sizeof(short));

	FEP(p)
		good = READCS(dcin, p->ai, sizeof(short));

	CHECK("Early EOF around player AI.");

	FEP(p)
		p->read_aifile(dcin);

	FEP(p)
		p->aimode = getc(dcin);

	if (!CHECKSECT(dcin, "player data 1"))
	{
		ret = ERR_data;
		strcpy(msg, e_data);
		goto Cleanup;
	}

	FEP(p)
		fread(&p->resources, sizeof(long), 6, dcin);

	/* Global Victory */

	CHECKSECT(dcin, "global victory");
	fread(&vict, sizeof(Victory), 1, dcin);

	/* Diplomacy */

	FEP(p)
		fread(p->diplomacy, sizeof(long), NUM_PLAYERS, dcin);

	fseek(dcin, 0x2D00, SEEK_CUR);	//should I check this? probably.

	/* Disables */

	CHECKSECT(dcin, "beginning of disables");

	UNREAD(dcin, sizeof(long) * NUM_PLAYERS);	//other allied victory

	//techs
	FEP(p)
		fread(&p->ndis_t, sizeof(long), 1, dcin);
	FEP(p)
		fread(&p->dis_tech, sizeof(long), perversion->max_disables1, dcin);	//60 in CC?

	//units
	FEP(p)
		fread(&p->ndis_u, sizeof(long), 1, dcin);
	FEP(p)
		fread(&p->dis_unit, sizeof(long), perversion->max_disables1, dcin);	//60 in CC?

	//buildings
	FEP(p)
		fread(&p->ndis_b, sizeof(long), 1, dcin);
	FEP(p)
		fread(&p->dis_bldg, sizeof(long), perversion->max_disables2, dcin);	//34 in CC?

	fseek(dcin, 0x8, SEEK_CUR);	//TEMP
	fread(&all_techs, sizeof(long), 1, dcin);
	FEP(p)
		fread(&p->age, sizeof(long), 1, dcin);

	/* Map */

	if (!CHECKSECT(dcin, "beginning of map"))
	{
		ret = ERR_data;
		strcpy(msg, e_data);
		goto Cleanup;
	}
	//this is so stupid (player 1's camera pos stored as longs)
	fread(&i, sizeof(long), 1, dcin);
	players[0].camera[1] = (float)i;
	fread(&i, sizeof(long), 1, dcin);
	players[0].camera[0] = (float)i;
	map.read(dcin, ver == VER_AOE2TC);

	/* Population Limits & Units */

	fread(&Player::num_players, sizeof(long), 1, dcin);

	for (i = PLAYER1_INDEX; i < GAIA_INDEX; i++)
	{
		UNREAD(dcin, sizeof(float) * PLAYER_POP_UNREAD_COUNT);	//2nd copy of resources
		if (ver == VER_AOE2TC)
			fread(&players[i].pop, 1, 4, dcin);
	}

	REPORTS(
		players[GAIA_INDEX].read_units(dcin),
		(ret = ERR_data, strcpy(msg, e_data)),
		"Scenario Data: GAIA unit read fail.\n"
		);

	for (i = PLAYER1_INDEX; i < 8 && ret == ERR_none; i++)
		players[i].read_units(dcin);

	/* Player Data 3 */

	UNREAD(dcin, sizeof(long));	//Number of players

	if (ret == ERR_none)
	for (i = PLAYER1_INDEX; i < 8; i++)
	{
		if (!players[i].read_data3(dcin,
			i == PLAYER1_INDEX ? editor_pos : NULL	//See scx_format.txt.
			))
		{
			ret = ERR_data;
			strcpy(msg, e_data);
			printf("Scenario Data: Player data 3 for player %d read fail.\n", i);
			goto Cleanup;
		}
	}
	UNKNOWN(double, 1.6, "post-PlayerData3 unknown");
	UNKNOWN(char, 0, "post-PlayerData3 unknown char");

	CHECKEOF();

	/* Triggers */

	unsigned int n_trigs;
	Trigger *t;
	good = (ret == ERR_none);	//check for error return before proceeding

	if (good)
	{
		fread(&n_trigs, sizeof(long), 1, dcin);
		triggers.allocate(n_trigs, true);
		t = triggers.first();
	}

	if (n_trigs && good)
	{
		i = n_trigs;

		//read triggers
		while (good && i--)
		{
			if (setts.intense)
				printf("Reading trigger %d.", n_trigs - i);
			good = t++->read(dcin);
		}

		CHECKEOF();

		CHECKA("Scenario Error: Trigger %d read fail.\n", n_trigs - i);

		//read trigger order
		t_order.allocate(n_trigs, true);
		fread(t_order.first(), sizeof(long), n_trigs, dcin);
		if (*t_order.first() > n_trigs)
		{
			ret = ERR_data;
			strcpy(msg, e_data);
			printf("Scenario Data: Trigger order list corrupted. Possibly out of sync?\n");
		}
	}

	/* Included Files */
	fread(&unk3, sizeof(long), 2, dcin);	//unk3, unk4

	if (unk3 == 1 && unk4 == 1)
		fseek(dcin, 396, SEEK_CUR);

	fread(&cFiles, sizeof(long), 1, dcin);
	if (cFiles > 0)
	{
		files = new AOKFile[cFiles];

		if (!files)
		{
			ret = ERR_mem;
			strcpy(msg, e_mem);
			goto Cleanup;
		}

		for (i = 0; i < cFiles; i++)
		{
			READCS(dcin, files[i].name, sizeof(long));
			files[i].data.read(dcin, sizeof(long));
		}
	}

Cleanup:

	fclose(dcin);
	printf("Done.\n");

	return ret;
}

/* Write to-be-compressed data to a temp file */

int Scenario::write_data(const char *path)
{
	FILE *dcout;
	int i, len;
	long num;
	float f;
	int num_players = Player::num_players;	//for quick access
	Player *p;

	dcout = fopen(path, "wb");

	/* Compressed header */

	fwrite(&next_uid, 4, 1, dcout);
	fwrite(&ver2, sizeof(float), 1, dcout);
	FEP(p)
	{
		len = strlen(p->name);
		fwrite(p->name, sizeof(char), len, dcout);
		NULLS(dcout, 0x100 - len);
	}
	FEP(p)
		fwrite(&p->stable, 4, 1, dcout);
	
	num = 0x00000004;
	FEP(p)
	{
		fwrite(&p->enable, sizeof(long), 3, dcout);
		fwrite(&num, 4, 1, dcout);
	}

	fwrite(&unknown, sizeof(long), 1, dcout);
	putc(0, dcout);
	f = -1;
	fwrite(&f, 4, 1, dcout);
	WRITECS(dcout, origname, false);

	/* Messages & Cinematics */
	if (perversion->mstrings)
		fwrite(mstrings, sizeof(long), perversion->messages_count, dcout);

	for (i = 0; i < perversion->messages_count; i++)
		messages[i].write(dcout, sizeof(short));

	for (i = 0; i < NUM_CINEM; i++)
	{
		WRITECS(dcout, cinem[i]);
	}

	fwrite(&bBitmap, 4, 1, dcout);
	fwrite(&bitmap.info_hdr.biWidth, 8, 1, dcout);	//width, height
	if (!bBitmap)
	{
		num = 1;
		fwrite(&num, 2, 1, dcout);
	}
	else
	{
		NEGS(dcout, 2);
		bitmap.write(dcout);
	}
	NULLS(dcout, 0x40);

	/* Player Data 2 */

	FEP(p)
	{
		WRITECS(dcout, p->ai, false);
	}
	FEP(p)
	{
		NULLS(dcout, 8);
		if (p->aifile)
		{
			len = strlen(p->aifile);
			fwrite(&len, 4, 1, dcout);
			fwrite(p->aifile, sizeof(char), len, dcout);
		}
		else
			NULLS(dcout, 4);	//0-length AI
	}
	FEP(p)
		putc(p->aimode, dcout);

	SECT(dcout);
	FEP(p)
		fwrite(&p->resources, sizeof(long), 6, dcout);

	/* Global Victory */

	SECT(dcout);
	fwrite(&vict, sizeof(vict), 1, dcout);

	/* Diplomacy */

#ifndef NOWRITE_D1
	FEP(p)
		fwrite(p->diplomacy, sizeof(long), NUM_PLAYERS, dcout);
#else
	NULLS(dcout, sizeof(long) * NUM_PLAYERS * NUM_PLAYERS);
#endif

	NULLS(dcout, 0x2D00);	//lol

	SECT(dcout);
	
	//2nd allied victory
	FEP(p)
	{
		num = p->avictory;
		fwrite(&num, sizeof(long), 1, dcout);
	}

	/* Disables */

	FEP(p)
		fwrite(&p->ndis_t, sizeof(long), 1, dcout);
	FEP(p)
		fwrite(&p->dis_tech, sizeof(long), perversion->max_disables1, dcout);
	FEP(p)
		fwrite(&p->ndis_u, sizeof(long), 1, dcout);
	FEP(p)
		fwrite(&p->dis_unit, sizeof(long), perversion->max_disables1, dcout);
	FEP(p)
		fwrite(&p->ndis_b, sizeof(long), 1, dcout);
	FEP(p)
		fwrite(&p->dis_bldg, sizeof(long), perversion->max_disables2, dcout);
	NULLS(dcout, 0x8);
	fwrite(&all_techs, sizeof(long), 1, dcout);
	FEP(p)
		fwrite(&p->age, 4, 1, dcout);

	/* Map */

	SECT(dcout);
	num = (long)players[0].camera[1];
	fwrite(&num, 4, 1, dcout);
	num = (long)players[0].camera[0];
	fwrite(&num, 4, 1, dcout);
	map.write(dcout, ver == VER_AOE2TC);

	/* Population Limits & Units */

	fwrite(&num_players, 4, 1, dcout);
	for (i = PLAYER1_INDEX; i < num_players - 1; i++)
	{
		float resources[6];
		resources[0] = (float)players[i].resources[2];	//food
		resources[1] = (float)players[i].resources[1];
		resources[2] = (float)players[i].resources[0];	//gold
		resources[3] = (float)players[i].resources[3];
		resources[4] = (float)players[i].resources[4];
		fwrite(resources, sizeof(float), PLAYER_POP_UNREAD_COUNT, dcout);

		if (ver == VER_AOE2TC)
			fwrite(&players[i].pop, 4, 1, dcout);
	}
	
	players[8].write_units(dcout);
	for (i = PLAYER1_INDEX; i < num_players - 1; i++)
	{
		players[i].write_units(dcout);
	}

	/* Player Data 3 */

	fwrite(&num_players, 4, 1, dcout);
	for (i = PLAYER1_INDEX; i < num_players - 1; i++)
	{
		players[i].write_data3(dcout, i,
			i == PLAYER1_INDEX ? editor_pos : NULL	//See scx_format.txt.
			);
	}
	double temp = 1.6;
	fwrite(&temp, 8, 1, dcout);
	putc(0, dcout);

	/* Triggers */

	Trigger *t_parse = triggers.first();
	num = triggers.count();
	fwrite(&num, sizeof(long), 1, dcout);

	i = num;
	while (i--)
	{
		t_parse->write(dcout);
		t_parse++;
	}

	fwrite(t_order.first(), sizeof(long), num, dcout);
	NULLS(dcout, 8);

	fclose(dcout);

	return ERR_none;
}

bool Scenario::export_bmp(const char *path)
{
	if (!bBitmap)
		return false;

	return bitmap.toFile(path);
}

#pragma warning(push)
#pragma warning(disable:4018)	//signed/unsigned mismatch necessary in this function

//I HATE THIS FUNCTION!
void Scenario::clean_triggers()
{
	size_t count = triggers.count();
	bool *kill = new bool[count];	//contains whether a trigger should be deleted
	bool *k_parse = kill;  //we pointer-parse this since the indexes get offset after deletion
	unsigned long *o_parse;
	Trigger *t_parse;

	//initialize array to true
	memset(kill, true, count);

	//save the ones that are still indexed on t_order
	count = t_order.count();
	o_parse = t_order.first();
	while (count--)
		kill[*o_parse++] = false;

	//now delete the ones still on the kill list
	count = triggers.count();
	for (index_t t_index = 0; t_index < count; t_index++)
	{
		if (!*k_parse++)	//If it shouldn't be deleted...
			continue;		//... skip all this

		int o_count = t_order.count();
		int t_count = count - 1;	//-1 for the trigger we're about to delete
		t_parse = triggers.first();
		o_parse = t_order.first();

		triggers.remove(t_index);

		//adjust t_order accordingly
		while (o_count--)
		{
			if (*o_parse > t_index)	//if this order index is beyond the one we just deleted
				(*o_parse)--;	//shift it forward

			o_parse++;
		}

		//adjust effect's trigger indexes accordingly (UGH!)
		while (t_count--)
		{
			int e_count = t_parse->effects.count();
			Effect *e_parse = t_parse->effects.first();

			while (e_count--)
			{
				if (e_parse->trig_index > t_index)
					e_parse->trig_index--;
				else if (e_parse->trig_index == t_index)
					e_parse->trig_index = (unsigned)-1;

				e_parse++;
			}

			t_parse++;
		}

		t_index--; count--;	  //update loop variables
	}

	delete [] kill;
}

#pragma warning(pop)

index_t Scenario::insert_trigger(Trigger *t, index_t after)
{
	index_t tindex;
	size_t count = scen.t_order.count();
	unsigned long *o_parse = t_order.first();

	tindex = scen.triggers.append(*t);

	if (after == -1)
		t_order.append(tindex);
	else
	{
		//scroll through the t_order list and insert the new item after the source
		while (count--)
		{
			if (*o_parse == after)
			{
				t_order.insert(o_parse + 1, tindex);
				break;
			}

			o_parse++;
		}
	}

	return tindex;
}

unsigned long *Scenario::find_trigger(unsigned long trigger)
{
	unsigned long *ret = NULL;
	int count = t_order.count();
	unsigned long *o_parse = t_order.first();

	while (count--)
	{
		if (*o_parse == trigger)
			ret = o_parse;

		o_parse++;
	}

	return ret;
}

bool Scenario::exFile(const char *directory, long index)
{
	int c;
	FILE *out;
	AOKFile *parse = scen.files;
	char buffer[_MAX_PATH];

	_mkdir(directory);
	
	if (index < 0)
	{
		c = scen.cFiles;
	}
	else
	{
		c = 1;
		parse += index;
	}

	while (c--)
	{
		sprintf(buffer, "%s\\%s.per", directory, parse->name);
		if (out = fopen(buffer, "wb"))
		{
			fwrite(parse->data.c_str(), sizeof(char), parse->data.length(), out);
			fclose(out);
		}
		else
			break;

		parse++;
	}

	return (c == -1);	//if it doesn't, we broke early
}

/*	MAP COPY

	Map copy data is stored as follows in memory:

	size_t width, height, unitcount[NUM_PLAYERS];
	Terrain terrain[x][y];
	Unit units[unitcount];	//x and y modified to dist from the bottom-left corner of copied area
*/

class MapCopyCache
{
public:
	RECT source;

	size_t unitcount[NUM_PLAYERS];
	SList<Unit> units;

	~MapCopyCache()
	{
		units.clear(true);
	}
};

#define ISINRECT(r, x, y) \
	(x >= r.left && x <= r.right && y >= r.bottom && y <= r.top)

int Scenario::map_size(const RECT &source, MapCopyCache *&mcc)
{
	Player *p;
	int ret;
	int i;
	Unit *u;
	size_t *unitcount;

	mcc = new MapCopyCache;

	CopyRect(&mcc->source, &source);

	/* x, u, unitcount */
	ret = sizeof(size_t) * 2;
	ret += sizeof(mcc->unitcount);

	/* space for the tiles */
	ret += sizeof(Map::Terrain) * (source.top - source.bottom + 1) * (source.right - source.left + 1);

	/* space for the units */
	unitcount = mcc->unitcount;
	FEP(p)
	{
		u = p->units.first();
		(*unitcount) = 0;

		for (int j = 0; j < p->units.count(); j++, u++)
		{
			if (ISINRECT(source, u->x, u->y))
			{
				ret += u->size;
				mcc->units.append(new Unit(*u));
				(*unitcount)++;
			}
		}

		unitcount++;
	}

	return ret;
}

AOKTS_ERROR Scenario::map_copy(Buffer &to, const MapCopyCache *mcc)
{
	size_t w, h;
	unsigned x, y;

	x = mcc->source.left;
	y = mcc->source.bottom;

	w = mcc->source.right - x;
	h = mcc->source.top - y;

	to.write(&w, sizeof(w));
	to.write(&h, sizeof(h));
	to.write(mcc->unitcount, sizeof(mcc->unitcount));

	map.writeArea(to, mcc->source);

	for (const SLink<Unit> *parse = mcc->units.first(); parse; parse = parse->next)
	{
		parse->item->x -= x;
		parse->item->y -= y;
		parse->item->toBuffer(to);
	}

	delete mcc;

	return ERR_none;
}

AOKTS_ERROR Scenario::map_paste(const POINT &to, Buffer &from)
{
	size_t x, y, unitcount[NUM_PLAYERS];
	RECT area;
	unsigned i;
	Player *p;

	from.read(&x, sizeof(x));
	from.read(&y, sizeof(y));
	from.read(unitcount, sizeof(unitcount));

	area.left = to.x;
	area.bottom = to.y;
	area.right = area.left + x;
	area.top = area.bottom + y;

	map.readArea(from, area);

	FEP(p)
	{
		/* find units in the paste area and erase them */
		for (int j = 0; j < p->units.count(); )
		{
			Unit *uparse = p->units.at(j);
			if (ISINRECT(area, uparse->x, uparse->y))
			{
				p->units.remove(uparse);
			}
			else
				j++;
		}

		/* add the units from the copy buffer */
		size_t count = unitcount[i];
		while (count--)
		{
			Unit u(from);
			u.ident = next_uid++;
			u.x += to.x;
			u.y += to.y;
			p->units.append(u);
		}
	}

	return ERR_none;
}

#ifdef OLD_MAP_COPY

int Scenario::map_copy(const RECT &source, const POINT &target)
{
	RECT rTarget, temp;
	int width, height, dx, dy;
	Map::Terrain *sparse, *tparse;	//source and target parsing
	int i;
	Player *p;
	int count;
	Unit *uparse;

	width = source.right - source.left;
	height = source.bottom - source.top;
	dx = target.x - source.left;
	dy = target.y - source.top;

	/* Check for overlapping (this may be supported in the future) */
	rTarget.left = target.x;
	rTarget.top = target.y;
	rTarget.right = target.x + width;
	rTarget.bottom = target.y + height;
	if (IntersectRect(&temp, &source, &rTarget))
		return ERR_overlap;

	/* Check for exceeding map boundaries */
	if ((unsigned)source.right > map.x || (unsigned)source.bottom > map.y ||
		(unsigned)rTarget.right > map.x || (unsigned)rTarget.bottom > map.y)
		return ERR_outofbounds;

	/* Terrain */
	for (int x = 0; x <= width; x++)	//inclusive
	{
		sparse = map.terrain[source.left + x] + source.top;
		tparse = map.terrain[target.x + x] + target.y;

		for (int y = 0; y <= height; y++)	//inclusive
		{
			tparse->cnst = sparse->cnst;
			tparse->elev = sparse->elev;
			sparse++;
			tparse++;
		}
	}

	/* Units. Believe me, it's more efficient this way. */
	FEP(p)
	{
		/*	not only is this count variable more efficient,
			it is essential so we don't remove added units */
		count = p->units.count();

		for (int j = 0; j < count; j++)
		{
			/*	I'd love to do an incremental parse here,
				but the vector might be reallocated. :-( */
			uparse = p->units.at(j);

			if (ISINRECT(rTarget, (int)uparse->x, (int)uparse->y))
			{
				p->units.remove(uparse);
			}
			else if (ISINRECT(source, (int)uparse->x, (int)uparse->y))
			{
				Unit u(*uparse);
				u.x += dx;
				u.y += dy;
				u.ident = next_uid++;
				p->units.append(u);
				uparse++;
			}
		}
	}

	return ERR_none;
}
#endif
int Scenario::getPlayerCount()
{
	int ret = 0;
	Player *p;
	int i;

	FEP(p)
		ret += p->enable;

	return ret;
}

/* Map */

Map::Terrain::Terrain()
:	cnst(0), elev(2)
{
}

Map::Map()
:	aitype(8), x(0), y(0)
{
}

void Map::reset()
{
	Terrain *parse;

	aitype = 8;	//8 is kinda the -1 for maps (since the base value is 9)
	x = *MapSizes;
	y = *MapSizes;

	parse = *terrain;
	for (int i = 0; i < MAX_MAPSIZE * MAX_MAPSIZE; i++)
	{
		parse->cnst = 0;
		parse->elev = 2;
		parse++;
	}
}

void Map::read(FILE *in, bool ex)
{
	Terrain *t_parse;
	unsigned int count;

	if (ex)
		fread(&aitype, 1, sizeof(long), in);	//aitype

	fread(&x, 2, sizeof(long), in);	//x, y

	for (unsigned int i = 0; i < x; i++)
	{
		count = y;
		t_parse = terrain[i];

		while (count--)
		{
			fread(t_parse++, sizeof(Terrain), 1, in);
			fgetc(in);	//skip the null
		}
	}
}

void Map::write(FILE *out, bool ex)
{
	unsigned int count;
	Terrain *t_parse;

	//there's no aitype in non-TC
	if (ex)
		fwrite(this, sizeof(long), 3, out);	//aitype, x, y
	else
		fwrite(&x, sizeof(long), 2, out);	//x, y

	for (unsigned int i = 0; i < x; i++)
	{
		count = y;
		t_parse = terrain[i];

		while (count--)
		{
			fwrite(t_parse++, sizeof(Terrain), 1, out);
			putc(0, out);
		}
	}
}

bool Map::readArea(Buffer &from, const RECT &area)
{
	Terrain *parse;

	for (unsigned i = area.left; i <= area.right; i++)
	{
		parse = terrain[i] + area.bottom;

		for (unsigned j = area.bottom; j <= area.top; j++)
		{
			from.read(parse, sizeof(*parse));
			parse++;
		}
	}

	return true;
}

bool Map::writeArea(Buffer &b, const RECT &area)
{
	Terrain *parse;

	/* perform some validation on input */
	if (area.bottom < 0 || area.top > y ||
		area.left < 0 || area.right > x)
		return false;

	for (unsigned i = area.left; i <= area.right; i++)
	{
		parse = terrain[i] + area.bottom;

		for (unsigned j = area.bottom; j <= area.top; j++)
		{
			b.write(parse, sizeof(Terrain));
			parse++;
		}
	}

	return true;
}

/* Units */

Unit::Unit(UID id)
:	x(0), y(0), u_float(1), ident(id), cnst(0),
	u_char(0), rotate(0), frame(0), garrison(-1),
	pType(NULL)
{}

Unit::Unit(Buffer &from)
{
	from.read(this, size);
	pType = esdata.getUnitById(cnst);
}

void Unit::read(FILE *in)
{
	fread(this, size, 1, in);
	pType = esdata.getUnitById(cnst);
}

const UnitLink *Unit::getType()
{
	return (pType) ? pType : (pType = (UnitLink*)getById(esdata.units, cnst));
}

const UnitLink *Unit::getType() const
{
	return (pType) ? pType : (UnitLink*)getById(esdata.units, cnst);
}

void Unit::setType(const UnitLink *t)
{
	assert(t);
	pType = t;
	cnst = (UCNST)t->id;
}

void Unit::toBuffer(Buffer &to) const
{
	to.write(this, size);
}

size_t Unit::size = 29;

/* Players */

int Player::num_players = 9;

const char *Player::names[] =
{
	"Player 1", "Player 2", "Player 3", "Player 4", "Player 5", "Player 6", "Player 7", "Player 8",
	"GAIA", "Player 10", "Player 11", "Player 12", "Player 13", "Player 14", "Player 15", "Player 16"
};

Player::Player()
:	aifile(NULL), stable(-1)
{
}

Player::~Player()
{
	if (aifile)
		delete [] aifile;
}

void Player::reset()
{
	memset(name, 0, 30);
	stable = -1;
	enable = human = false;
	civ = 1;	//TODO: ini var
	strcpy(ai, "RandomGame");
	aimode = AI_standard;
	if (aifile)
	{
		delete [] aifile;
		aifile = NULL;
	}
	memset(resources, 0, sizeof(resources));
	ndis_t = ndis_b = ndis_u = 0;
	memset(dis_tech, 0, sizeof(dis_tech));
	memset(dis_unit, 0, sizeof(dis_unit));
	memset(dis_bldg, 0, sizeof(dis_bldg));
	age = 0;
	pop = 75;
	camera[0] = 0;
	camera[1] = 0;
	avictory = false;
	memset(diplomacy, 0, sizeof(diplomacy));
	color = 0;
	ucount = 2;
	units.clear();
}

bool Player::read_aifile(FILE *in)
{
	int len;

	fseek(in, 8, SEEK_CUR);
	fread(&len, 1, 4, in);
	if (len > 0)
	{
		aifile = new char[len + 1];
		if (aifile == NULL)
			return false;

		fread(aifile, sizeof(char), len, in);
		aifile[len] = '\0';
	}

	return true;
}

bool Player::read_units(FILE *in)
{
	unsigned long num;
	bool ret = true;
	int i;
	Unit u;

	fread(&num, 1, sizeof(long), in);
	if (num &&
		(ret = units.allocate(num, false))
		)
	{
		i = num;
		while (i--)
		{
			u.read(in);
			units.append(u);
		}
	}
	else if (!ret)	//naff.
		printf("Unit read fail: could not allocate space for %d units.\n", num);

	return ret;
}

void Player::write_units(FILE *out)
{
	unsigned int num = units.count();
	Unit *u;

	fwrite(&num, sizeof(long), 1, out);
	u = units.first();
	while (num--)
		fwrite(u++, Unit::size, 1, out);
}

bool Player::read_data3(FILE *in, float *view)
{
	bool ret = true;
	int len = 0;
	long end;
	char unk;

	fread(&len, sizeof(short), 1, in);
	UNREAD(in, len);

	fread(
		view ? view : camera,	//use view[] if calling func wants us to store there.
		sizeof(float), 2, in);

	fread(&u1, sizeof(short), 2, in);	//u1, u2
	avictory = (fgetc(in) == 1);
	UNREAD(in, sizeof(short));	//playercount
	UNREAD(in, sizeof(char) * num_players);	//diplomacy
	UNREAD(in, sizeof(long) * num_players);	//diplomacy2
	fread(&color, sizeof(long), 2, in);	//color, ucount

	REPORTS(ucount == 1 || ucount == 2, ret = false,
		"Scenario Data: Unknown Float value unrecognized. Don\'t know what to expect next.\n");

	fread(&unk, sizeof(char), 1, in);
	if (unk)
		fseek(in, 44, SEEK_CUR);	//gte hack.

	if (ucount == 2)
		fseek(in, 8, SEEK_CUR);

	fseek(in, 8, SEEK_CUR);
	fread(&end, sizeof(long), 1, in);

	return ret;
}

void Player::write_data3(FILE *out, int me, float *view)
{
	int d_convert[4] = { 2, 3, 1, 4 };
	int i;

	WRITECS(out, Player::names[me]);

	fwrite(
		view ? view : camera,	//let calling func determine view if it wants to
		sizeof(float), 2, out);	//camera[]

	fwrite(&u1, sizeof(short), 2, out);	//u1, u2

	fputc(avictory, out);
	fwrite(&num_players, sizeof(short), 1, out);

#ifndef NOWRITE_D2
	putc(diplomacy[GAIA_INDEX], out);	//GAIA first
	for (i = 0; i < num_players - 1; i++)
		putc(diplomacy[i], out);
#else
	NULLS(out, num_players);
#endif

#ifndef NOWRITE_D3
	NULLS(out, 4);	//gaia diplomacy
	for (i = 0; i < num_players - 1; i++)
	{
		if (i == me)
			fwrite(&d_convert[2], sizeof(long), 1, out);
		else
			fwrite(&d_convert[diplomacy[i]], sizeof(long), 1, out);
	}
#else
	NULLS(out, num_players * sizeof(long));
#endif

	fwrite(&color, 4, 2, out);	//color, ucount
	NULLS(out, 9);
	if (ucount == 1)
	{
		NULLS(out, 4);
	}
	else
	{
		NULLS(out, 8);
		NEGS(out, 4);
	}
}

bool Player::import_ai(const char *path)
{
	FILE *ai_in;
	size_t ai_len = fsize(path);
	size_t read_len;

	if (aifile)
		delete [] aifile;

	aifile = new char[ai_len + 1];

	ai_in = fopen(path, "rb");
	if (!ai_in)
		return false;

	read_len = fread(aifile, sizeof(char), ai_len, ai_in);

	aifile[ai_len] = '\0';	//null-terminate it

	return read_len == ai_len;
}


bool Player::export_ai(const char *path)
{
	FILE *ai_out;

	//check that player actually has an ai script.
	if (!*aifile)
		return false;

	ai_out = fopen(path, "wb");
	if (!ai_out)
		return false;

	fputs(aifile, ai_out);

	fclose(ai_out);
	return true;
}

