#include <stdio.h>

#include "trigger.h"
#include "utilio.h"

/* Triggers, Conditions, and Effects */

#define EFFECT_MEMBERS 22	//count of all the static members of class Effect

#undef PLAYER1_INDEX
#undef GAIA_INDEX
#define PLAYER1_INDEX 1
#define GAIA_INDEX 0

ECBase::ECBase(enum TType c, long t)
:	type(t), ttype(c)
{}

Effect::Effect()
:	ECBase(EFFECT)
{
	clear();
}

Effect::Effect(char *data, int size)
:	ECBase(EFFECT)
{
	Buffer buffer(data, size);
	frombuffer(buffer);
}

void Effect::clear()
{
	snd_len = 0;
	memset(&ai_goal, -1, EFFECT_MEMBERS * sizeof(long));	//ouch
	pUnit = NULL;
	pTech = NULL;

	memset(sound, 0, sizeof(sound));
	memset(uids, -1, sizeof(uids));
}

void Effect::tobuffer(Buffer &buffer) const
{
	buffer.write(&ttype, sizeof(long));
	buffer.write(&type, sizeof(long));
	buffer.write(&ai_goal, 6 * sizeof(long));
	buffer.writel(pUnit);
	buffer.write(&s_player, 2 * sizeof(long));
	buffer.writel(pTech);
	buffer.write(&stringid, 13 * sizeof(long));
	buffer.writes(text, 4);
	buffer.writes(sound, sizeof(long));
	if (num_sel > 0)
		buffer.write(uids, sizeof(uids));
}

void Effect::frombuffer(Buffer &buffer)
{
	buffer.read(&ttype, sizeof(long));
	buffer.read(&type, sizeof(long));
	buffer.read(&ai_goal, 6 * sizeof(long));
	buffer.readl((const Link*&)pUnit, esdata.units);
	buffer.read(&s_player, 2 * sizeof(long));
	buffer.readl((const Link*&)pTech, esdata.techs);
	buffer.read(&stringid, 13 * sizeof(long));
	buffer.reads(text, sizeof(long));
	buffer.reads(sound, sizeof(long));
	snd_len = strlen(sound) + 1;
	if (num_sel > 0)
		buffer.read(uids, sizeof(uids));
}

void Effect::set(const Effect &e)
{
	type = e.type;
	memcpy(&ai_goal, &e.ai_goal, EFFECT_MEMBERS * sizeof(long));
	memcpy(sound, &e.sound, sizeof(sound));
	if (e.num_sel > 0)
		memcpy(uids, &e.uids, sizeof(uids));
	snd_len = e.snd_len;

	text.set(e.text.c_str());

	pTech = e.pTech;
	pUnit = e.pUnit;
}

bool Effect::read(FILE *in)
{
	long temp;

	fread(&type, 8, sizeof(long), in);
	fread(&temp, 1, sizeof(long), in);
	pUnit = (UnitLink*)getById(esdata.units, temp);
	fread(&s_player, 2, sizeof(long), in);
	fread(&temp, 1, sizeof(long), in);
	pTech = (TechLink*)getById(esdata.techs, temp);
	fread(&stringid, 13, sizeof(long), in);

	if (ttype != EFFECT)
	{
		return false;
	}

	text.read(in, sizeof(long));

	fread(&snd_len, sizeof(long), 1, in);
	if (snd_len > 0)
	{
		fread(sound, sizeof(char), snd_len, in);
		sound[snd_len] = '\0';
	}

	if (num_sel > 0)
		fread(uids, sizeof(long), num_sel, in);

	if (type > MAX_EFFECT)
		printf("WARNING: Unknown effect %d.\n", type);

	return true;
}

void Effect::write(FILE *out)
{
	long temp;

	fwrite(&type, sizeof(long), 8, out);
	temp = (pUnit) ? pUnit->id : -1;
	fwrite(&temp, sizeof(long), 1, out);
	fwrite(&s_player, sizeof(long), 2, out);
	temp = (pTech) ? pTech->id : -1;
	fwrite(&temp, sizeof(long), 1, out);
	fwrite(&stringid, sizeof(long), 13, out);
	text.write(out, sizeof(long));
	fwrite(&snd_len, sizeof(long), 1, out);
	fwrite(sound, sizeof(char), snd_len, out);
	if (num_sel > 0)
		fwrite(uids, sizeof(long), num_sel, out);
}

int Effect::size() const
{
	int ret = 26 * sizeof(long) + sizeof(sound) + sizeof(uids);
	ret += text.total_size();
	return ret;
}

bool Effect::check() const
{
	switch (type)
	{
	case EFFECT_ChangeDiplomacy:
		return (s_player >= 0 && t_player >= 0 && diplomacy >= 0);

	case EFFECT_ResearchTechnology:
		return (s_player >= 0 && pTech);

	case EFFECT_SendChat:
		return (s_player >= 0 && *text.c_str());	//AOK missing text check

	case EFFECT_PlaySound:
		return (s_player >= 0 && *sound);	//AOK missing sound check

	case EFFECT_SendTribute:
		return (s_player >= 0 && t_player >= 0 && res_type >= 0);

	case EFFECT_UnlockGate:
	case EFFECT_LockGate:
		return (num_sel >= 0);

	case EFFECT_ActivateTrigger:
	case EFFECT_DeactivateTrigger:
		return (trig_index >= 0);

	case EFFECT_AIScriptGoal:
		return (s_player >= 0 && ai_goal >= 0);

	case EFFECT_CreateObject:
		return (s_player >= 0 &&
			location.x >= 0 && location.y >= 0 && pUnit);

	case EFFECT_TaskObject:
	case EFFECT_KillObject:
	case EFFECT_RemoveObject:
	case EFFECT_FreezeUnit:
	case EFFECT_StopUnit:
		return (num_sel >= 0 || area.left >= 0 || utype >= 0);	//AOK missing this

	case EFFECT_DeclareVictory:
		return (s_player >= 0);

	//EFFECT_KillObject, EFFECT_RemoveObject above.

	case EFFECT_ChangeView:
	case EFFECT_SnapView:
		return (s_player >= 0 && location.x >= 0 && location.y >= 0);

	case EFFECT_Unload:
		return (s_player >= 0
			&& (num_sel >= 0 || area.left >= 0 || utype >= 0)	//AOK missing this
			&& location.x >=0 && location.y >= 0);

	case EFFECT_ChangeOwnership:
		return (s_player >= 0 && t_player >= 0
			&& (num_sel >= 0 || area.left >= 0 || utype >= 0));	//AOK missing this

	case EFFECT_Patrol:
		return (num_sel >= 0 && location.x >= 0 && location.y >= 0);

	case EFFECT_DisplayInstructions:
		return (panel >= 0 && disp_time >= 0
			&& (*text.c_str() || stringid));	//AOK missing text

	case EFFECT_ClearInstructions:
		return (panel >= 0);

	//EFFECT_FreezeUnit above

	case EFFECT_UseAdvancedButtons:
		return true;	//no properties to set

	case EFFECT_DamageObject:
	case EFFECT_ChangeObjectHP:
	case EFFECT_ChangeObjectAttack:
		return (amount != 0		//amount can be negative
			&& (num_sel >= 0 || area.left >= 0));	//AOK missing this

	case EFFECT_PlaceFoundation:
		return (s_player >= 0 && pUnit
			&& location.x >= 0 && location.y >= 0);

	case EFFECT_ChangeObjectName:
		return (num_sel >= 0);	//no text check

	//EFFECT_ChangeObjectHP, EFFECT_ChangeObjectAttack, EFFECT_StopUnit above

	//EFFECT_SnapView shares with EFFECT_ChangeView above

	case EFFECT_EnableTech:
	case EFFECT_DisableTech:
		return (pTech != NULL);

	case EFFECT_EnableUnit:
	case EFFECT_DisableUnit:
		return (pUnit != NULL);

	case EFFECT_FlashObjects:
		return (area.left >= 0); //SWGB missing this

	default:
		return false;	//unknown effect type
	}
}

#define CONDITION_MEMBERS 16

Condition::Condition()
:	ECBase(CONDITION)
{
	clear();
}

Condition::Condition(char *data, int size)
:	ECBase(CONDITION)
{
	Buffer b(data, size);
	frombuffer(b);
}

void Condition::clear()
{
	memset(&amount, -1, CONDITION_MEMBERS * sizeof(long));	//ouch again
	pTech = NULL;
	pUnit = NULL;
}

int Condition::size() const
{
	return 18 * sizeof(long);
}

bool Condition::check() const
{
	switch (type)
	{
	case CONDITION_BringObjectToArea:
		return (object >= 0 && area.left >= 0);

	case CONDITION_BringObjectToObject:
		return (object >= 0 && u_loc >= 0);

	case CONDITION_OwnObjects:
	case CONDITION_OwnFewerObjects:
	case CONDITION_OwnFewerFoundations:
		return (player >= 0 && amount >= 0);

	case CONDITION_ObjectsInArea:
	case CONDITION_SelectedObjectsInArea:
	case CONDITION_PoweredObjectsInArea:
		return (area.left >= 0 && amount >= 0);

	case CONDITION_DestroyObject:
		return (object >= 0);

	case CONDITION_CaptureObject:
		return (object >= 0 && player >= 0);

	case CONDITION_AccumulateAttribute:
		return (player >= 0 && res_type >= 0);

	case CONDITION_ResearchTehcnology:
	case CONDITION_ResearchingTechnology:
		return (player >= 0 && pTech->id >= 0);

	case CONDITION_Timer:
		return (timer >= 0);

	case CONDITION_ObjectSelected:
		return (object >= 0);

	case CONDITION_AISignal:
		return (ai_signal >= 0);

	case CONDITION_PlayerDefeated:
	case CONDITION_UnitsQueuedPastPopCap:
		return (player >= 0);

	case CONDITION_ObjectHasTarget:
		return (object >= 0 && u_loc >= 0);

	case CONDITION_ObjectVisible:
	case CONDITION_ObjectNotVisible:
		return (object >= 0);

	//CONDITION_ResearchingTechnology above

	case CONDITION_UnitsGarrisoned:
		return (object >= 0 && amount >= 0);

	case CONDITION_DifficultyLevel:
		return (amount >= 0);

	//CONDITION_OwnFewerFoundations shares with OwnObjects above
	//CONDITION_SelectedObjectsInArea shares with ObjectsInArea above
	//CONDITION_PoweredObjectsInArea shares with ObjectsInArea above
	//CONDITION_UnitsQueuedPastPopCap shares with PlayerDefeated above

	default:
		return false;
	}
}

bool Condition::read(FILE *in)
{
	long temp;

	fread(&type, sizeof(long), 6, in);
	fread(&temp, sizeof(long), 1, in);
	if (temp >= 0)
		pUnit = (UnitLink*)getById(esdata.units, temp);

	fread(&player, sizeof(long), 1, in);

	fread(&temp, sizeof(long), 1, in);
	if (temp >= 0)
		pTech = (TechLink*)getById(esdata.techs, temp);

	fread(&timer, sizeof(long), 9, in);

	if (type > MAX_CONDITION)
		printf("WARNING: Unknown condition %d.\n", type);

	return (ttype == CONDITION);
}

void Condition::write(FILE *out)
{
	long temp;

	fwrite(&type, sizeof(long), 6, out);
	temp = (pUnit) ? pUnit->id : -1;
	fwrite(&temp, sizeof(long), 1, out);
	fwrite(&player, sizeof(long), 1, out);
	temp = (pTech) ? pTech->id : -1;
	fwrite(&temp, sizeof(long), 1, out);
	fwrite(&timer, sizeof(long), 9, out);
}

void Condition::frombuffer(Buffer &b)
{
	b.read(&ttype, sizeof(long));
	b.read(&type, sizeof(long));
	b.read(&amount, sizeof(long) * 4);
	b.readl((const Link*&)pUnit, esdata.units);
	b.read(&player, sizeof(long));
	b.readl((const Link*&)pTech, esdata.techs);
	b.read(&timer, sizeof(long) * 9);
}

void Condition::tobuffer(Buffer &b) const
{
	b.write(&ttype, sizeof(long));
	b.write(&type, sizeof(long));
	b.write(&amount, sizeof(long) * 4);
	b.writel(pUnit);
	b.write(&player, sizeof(long));
	b.writel(pTech);
	b.write(&timer, sizeof(long) * 9);
}

Trigger::Trigger()
:	state(1), loop(0), u1(0), obj(0), obj_order(0)
{
	memset(name, 0, sizeof(name));
}

Trigger::Trigger(const Trigger &t)
:	state(t.state), loop(t.loop), u1(t.u1), obj(t.obj), obj_order(t.obj_order),
	description(t.description), effects(t.effects), conds(t.conds)
{
	strcpy(name, t.name);
}

Trigger::Trigger(char *data, const int size)
{
	int i, num;
	Buffer buffer(data, size);
	Effect *effects;
	Condition *conds;

	buffer.read(&state, 14);
	buffer.fill(0, sizeof(long));
	buffer.reads(description, sizeof(long));
	buffer.reads(name, sizeof(long));

	/* effects */

	buffer.read(&num, sizeof(long));
	if (!(effects = new Effect[num]))
		return;

	for (i = 0; i < num; i++)
		(effects + i)->frombuffer(buffer);

	while (num--)
	{
		buffer.read(&i, sizeof(long));
		this->effects.append(effects[i]);
	}

	delete [] effects;

	/* conditions */

	buffer.read(&num, sizeof(long));
	if (!(conds = new Condition[num]))
		return;

	for (i = 0; i < num; i++)
		(conds + i)->frombuffer(buffer);

	while (num--)
	{
		buffer.read(&i, sizeof(long));
		this->conds.append(conds[i]);
	}

	delete [] conds;
}

bool Trigger::read(FILE *in)
{
	bool ret = true;
	long n_effects, n_conds;
	Effect *effects = NULL;
	Condition *conditions = NULL;
	int i;

	//check for eof before starting
	REPORT(!feof(in), return false);

	fread(this, sizeof(char), 14, in);
	fseek(in, sizeof(long), SEEK_CUR);
	description.read(in, sizeof(long));
	NREADCS(in, name, sizeof(name), sizeof(long));

	//read effects
	fread(&n_effects, 1, sizeof(long), in);

	if (n_effects > 0)
	{
		if (!(effects = new Effect[n_effects]))
		{
			printf("Error: Failed to allocate memory.");
			goto Cleanup;
		}

		for (i = 0; i < n_effects; i++)
		{
			if (!(effects + i)->read(in))
			{
				ret = false;
				printf("Effect %d invalid.\n", i);
				goto Cleanup;
			}
		}

		while (n_effects--)
		{
			long order;

			fread(&order, 1, sizeof(long), in);
			this->effects.append(effects[order]);	//ordered in memory
		}
	}

	//read conditions
	fread(&n_conds, 1, sizeof(long), in);

	if (n_conds)
	{
		if (!(conditions = new Condition[n_conds]))
		{
			printf("Error: Failed to allocate memory.");
			goto Cleanup;
		}

		for (i = 0; i < n_conds; i++)
		{
			if (!(conditions + i)->read(in))
			{
				ret = false;
				printf("Condition %d invalid.\n", i);
				goto Cleanup;
			}
		}

		while (n_conds--)
		{
			long order;

			fread(&order, 1, sizeof(long), in);
			conds.append(conditions[order]);
		}
	}

Cleanup:

	delete [] effects;
	delete [] conditions;

	return ret;
}

void Trigger::write(FILE *out)
{
	int num, i;

	fwrite(&state, 14, 1, out);	//state, loop, u1, obj, obj_order
	NULLS(out, 4);
	description.write(out, sizeof(long));
	num = strlen(name) + 1;
	fwrite(&num, 4, 1, out);
	fwrite(name, sizeof(char), num, out);

	//Effects

	num = effects.count();

	fwrite(&num, 4, 1, out);
	for (i = 0; i < num; i++)
		effects.at(i)->write(out);

	for (i = 0; i < num; i++)
		fwrite(&i, 4, 1, out);

	//Conditions

	num = conds.count();

	fwrite(&num, 4, 1, out);
	for (i = 0; i < num; i++)
		conds.at(i)->write(out);

	for (i = 0; i < num; i++)
		fwrite(&i, 4, 1, out);
}

int Trigger::size() const
{
	int total = 0, num;

	total += 18;	//first members
	total += sizeof(long) + description.length() + 1;
	total += sizeof(long) + strlen(name) + 1;

	num = effects.count();

	total += sizeof(long);	//effect count

	for (int i = 0; i < num; i++)
		total += effects.at(i)->size();

	total += num * sizeof(long);	//order

	num = conds.count();

	total += sizeof(long);	//condition count
	total += sizeof(long) * 18 * num;	//conditions
	total += sizeof(long) * num;	//order

	return total;
}

void Trigger::tobuffer(char *dest, int size) const
{
	int i, num;
	Buffer buffer(dest, size);

	buffer.write(&state, 14);
	buffer.fill(0, sizeof(long));
	buffer.writes(description, sizeof(long));
	buffer.writes(name, sizeof(long));

	//effects

	num = effects.count();

	buffer.write(&num, sizeof(long));
	for (i = 0; i < num; i++)
		effects.at(i)->tobuffer(buffer);

	for (i = 0; i < num; i++)
		buffer.write(&i, sizeof(long));

	//conditions

	num = conds.count();

	buffer.write(&num, sizeof(long));

	for (i = 0; i < num; i++)
		conds.at(i)->tobuffer(buffer);

	//condition order
	for (i = 0; i < num; i++)
		buffer.write(&i, sizeof(long));
}

int Trigger::get_player()
{
	int ret = 0;
	int count;
	Effect *e;
	Condition *c;

	FOREACH(effects, e)
	{
		if (e->s_player > GAIA_INDEX && e->s_player != ret)
		{
			if (!ret)
				ret = e->s_player;
			else
			{
				ret = -1;
				break;
			}
		}
	}

	if (ret != -1)
	FOREACH(conds, c)
	{
		if (c->player > GAIA_INDEX && c->player != ret)
		{
			if (!ret)
				ret = c->player;
			else
			{
				ret = -1;
				break;
			}
		}
	}

	return ret;
}

void Trigger::change_player(unsigned to)
{
	int count;
	Effect *e;
	Condition *c;

	FOREACH(effects, e)
	{
		if (e->s_player > GAIA_INDEX)
			e->s_player = to;
	}

	FOREACH(conds, c)
	{
		if (c->player > GAIA_INDEX)
			c->player = to;
	}
}

const char *Effect::types[] =
{
	"Undefined",
	"Change Diplomacy",
	"Research Technology",
	"Send Chat",
	"Play Sound",
	"Send Tribute",
	"Unlock Gate",
	"Lock Gate",
	"Activate Trigger",
	"Deactivate Trigger",
	"AI Script Goal",
	"Create Object",
	"Task Object",
	"Declare Victory",
	"Kill Object",
	"Remove Object",
	"Change View",
	"Unload",
	"Change Ownership",
	"Patrol",
	"Display Instructions",
	"Clear Instructions",
	"Freeze Unit",
	"Use Advanced Buttons",
	"Damage Object",
	"Place Foundation",
	"Change Object Name",
	"Change Object HP",
	"Change Object Attack",
	"Stop Unit",
	"Snap View",
	"Unknown (31)",
	"Enable Tech",
	"Disable Tech",
	"Enable Unit",
	"Disable Unit",
	"Flash Objects"
};

const char *Condition::types[] =
{
	"Undefined",
	"Bring Object to Area",
	"Bring Object to Object",
	"Own Objects",
	"Own Fewer Objects",
	"Objects in Area",
	"Destroy Object",
	"Capture Object",
	"Accumulate Attribute",
	"Research Technology",
	"Timer",
	"Object Selected",
	"AI Signal",
	"Player Defeated",
	"Object Has Target",
	"Object Visible",
	"Object Not Visible",
	"Researching Technology",
	"Units Garrisoned",
	"Difficulty Level",
	"Own Fewer Foundations",
	"Selected Objects in Area",
	"Powered Objects in Area",
	"Units Queued Past Pop Cap"
};
