/**
	AOK Trigger Studio (See aokts.cpp for legal conditions.)
	WINDOWS VERSION.
	utilio.h -- File i/o Utility Functions
**/

#include <stdio.h>
#include <string.h>
#include <assert.h>

/*
	REPORT(expr, action)
	
	If [expr] evaluates false, performs [action].
	Also brings up a debug box in a debug build.
*/
#ifdef _DEBUG

#define REPORT(expr, action) \
	if (!(expr)) \
	{ \
		assert(false); \
		action; \
	}

#define REPORTS(expr, action, string) \
	if (!(expr)) \
	{ \
		assert(false); \
		action; \
		printf(string); \
	}

#else

#define REPORT(expr, action) \
	if (!(expr)) \
		action;

#define REPORTS(expr, action, string) \
	if (!(expr)) \
	{ \
		action; \
		printf(string); \
	}

#endif

//Quick Filesize
int fsize(const char *path);

/* Input */

//read from a FILE stream to a char *
inline bool READCS(FILE *in, char *dest, size_t lensize)
{
	size_t len = 0;	//set to zero for non-replaced bytes
	bool ret;

	fread(&len, 1, lensize, in);
	ret = fread(dest, sizeof(char), len, in) == len;
	dest[len] = '\0';

	return ret;
}

//read from a FILE stream to a char*, with a buffer size
inline bool NREADCS(FILE *in, char *dest, size_t space, size_t lensize)
{
	size_t len = 0;	//set to zero for non-replaced bytes
	bool ret;

	fread(&len, 1, lensize, in);

	if (len <= space)
	{
		ret = fread(dest, sizeof(char), len, in) == len;
		dest[len] = '\0';
	}
	else
	{
		fread(dest, sizeof(char), space, in);
		fseek(in, len - space, SEEK_CUR);
		dest[space] = '\0';
	}

	return ret;
}

//check the 0xFFFFFF9D separator
inline bool CHECKSECT(FILE *in, const char *section)
{
	bool ret = true;
	long sect;

	fread(&sect, sizeof(long), 1, in);

	REPORT(sect == 0xFFFFFF9D,
		(ret = false, printf("Out of sync around %s (%x).\n", section, ftell(in))));

#ifdef _DEBUG
	printf("Successfully passed section check for %s at %x.\n", section, ftell(in));
#endif

	fflush(stdout);

	return ret;
}

/* Output */

//write a variable amount of NULLs to an ostream
inline void NULLS(FILE *out, int size)
{
	while (size--)
		putc(0, out);
}

//write a variable amount of 0xFFs
inline void NEGS(FILE *out, int count)
{
	while (count--)
		putc(0xFF, out);
}

//write 0xFFFFFF9D
inline void SECT(FILE *out)
{
	long sect = 0xFFFFFF9D;
	fwrite(&sect, sizeof(long), 1, out);
}

//write a char*
inline void WRITECS(FILE *out, const char *source, bool null = true)
{
	int len = strlen(source);
	if (len)
		len += null;	//only write if string exists (mimicing AOK)

	fwrite(&len, sizeof(short), 1, out);
	fwrite(source, sizeof(char), len, out);
}
