/**
	AOK Trigger Studio (See aokts.cpp for legal conditions.)
	WINDOWS VERSION.
	utilui.h -- Custom controls and utility UI functions & macros.
**/

/* Shared msgbox strings */

const char errorOpenClipboard[] =
"Could not open clipboard. Most likely another window is currently using it.";
const char errorSetClipboard[] =
"Could not set clipboard data.";
const char errorAllocFailed[] =
"Could not allocate the memory for this operation. This could be a sign that you are out of memory, "
"but usually indicates a bug. Please contact DiGiT about this error.";
const char warningNoFormat[] =
"Specified clipboard data did not exist.\nPlease report this error to cyan.spam@gmail.com.";


/* Common functions */

void RegisterControls(HINSTANCE instance);

/* LCOMBO */

//see resource.h for custom styles

enum LCOMBO_msgs
{
	CBL_FILL = WM_APP,
	/*	Clears contents and fills the combobox with the provided items, selecting
		the item specified by lParam.
	
		wParam: Link *sel_link
		lParam: Link *list;
		return: index of selection
		*/
#define LCombo_Fill(d, i, s, l) \
	SendDlgItemMessage(d, i, CBL_FILL, (WPARAM)s, (LPARAM)l);

	CBL_FILLBYID,
	/*	Clears contents and fills the combobox with the provided items, selecting
		the item specified by lParam.
	
		wParam: int sel_id
		lParam: Link *list;
		return: index of selection
		*/
#define LCombo_FillById(d, i, s, l) \
	SendDlgItemMessage(d, i, CBL_FILL, (WPARAM)s, (LPARAM)l);

	CBL_GETSELID,
	/*	wParam: unused
		lParam: unused
		return: Link::id of current selection
		*/
#define LCombo_GetSelId(d, i) \
	SendDlgItemMessage(d, i, CBL_GETSELID, 0, 0);

	CBL_GETSELPTR,
	/*	wParam: unused
		lParam: unused
		return: Link* of current selection
		*/
#define LCombo_GetSelPtr(d, i) \
	SendDlgItemMessage(d, i, CBL_GETSELPTR, 0, 0);

	CBL_FIND,
	/*	wParam: unused
		lParam: data of desired item
		return: index of deisred item
		*/
#define LCombo_Find(d, i, x) \
	SendDlgItemMessage(d, i, CBL_FIND, 0, (LPARAM)x);

	CBL_SELBYID,
	/*	wParam: the Link::id to select.
		lParam: unused
		return: index of selected item
		*/
#define LCombo_SelById(d, i, x) \
	SendDlgItemMessage(d, i, CBL_SELBYID, x, 0);
};

#define ENABLE_WND(id, en) EnableWindow(::GetDlgItem(dialog, id), en)

/* Fills a combobox with an array of strings. Dialog HWND must be "dialog" in func. */
#define FILL_CB(id, array, count) \
	control = GetDlgItem(dialog, id); \
	for (i = 0; i < count; i++) \
		SendMessage(control, CB_ADDSTRING, 0, (LPARAM)array[i]);

/*
	FillCB: Fills a combobox with strings with associated data and selects one.
*/
void FillCB(HWND combobox, unsigned count, const struct PAIR *pairs, unsigned select);

/* Gets the current selection of a combo box, then its item-data */ 
#define GETSELDATA(id) \
	(control = GetDlgItem(dialog, id),\
	SendMessage(control,\
		CB_GETITEMDATA,\
		SendMessage(control, CB_GETCURSEL, 0, 0),\
		0))

/* Gets the current selection of a combo box, then its item-data */ 
#define GETLBSELDATA(id) \
	(control = GetDlgItem(dialog, id),\
	SendMessage(control,\
		LB_GETITEMDATA,\
		SendMessage(control, LB_GETCURSEL, 0, 0),\
		0))

/*
	CB_getid: Gets the id from a linkedlist list box
*/
inline long LB_getid(HWND listbox)
{
	int ret = SendMessage(listbox, LB_GETITEMDATA,
		SendMessage(listbox, LB_GETCURSEL, 0, 0), 0);
	
	return (ret != LB_ERR) ? ((Link*)ret)->id : -1;
}

#ifdef SSTRING_H
/*
	GetWindowText: Retrieves text from a window and stores it in an SString class.
*/
void GetWindowText(HWND wnd, SString &value);
#endif

/*
	GetDlgItemFloat: Return a float from a dialog box edit control.

	dialog: The owner dialog.
	id: The identifier of the control.

  Note: Text longer than 32 characters will be truncated.
*/
float GetDlgItemFloat(HWND dialog, UINT id);

/*
	SetDlgItemFloat: Set a dialog box edit control's text to a floating-point value.

	dialog: The owner dialog.
	id: The identifier of the control.
	value: The value to display.
*/
void SetDlgItemFloat(HWND dialog, UINT id, float value);

/*
	GetCheckedRadio: Returns the ID of the checked radio button from a range of radio buttons.
*/
UINT GetCheckedRadio(HWND dialog, UINT first, UINT last);

inline int get_index(const int *list, int value, int count)
{
	for (int i = 0; i < count; i++)
	{
		if (*list++ == value)
			return i;
	}

	return -1;
}


#ifdef WC_TREEVIEW

/*
	Misc Treeview Inline Funcs

  Note: These apply to all treeviews, not just this trigger tree.
*/

//Get the lParam of a specified item. If which = NULL, get the current selected item.
LPARAM GetItemParam(HWND treeview, HTREEITEM which);

/*
	TreeView_AddChild:
		Adds a child node to the specified parent with an
		lParam of param.
*/
void TreeView_AddChild(HWND treeview, LPARAM param, HTREEITEM parent, HTREEITEM after);

#define IS_ROOTNODE(treeview, item) \
	(SendMessage(treeview, TVM_GETNEXTITEM, TVGN_PARENT, (LPARAM)item) == 0)

//Get the "selected" root item.
inline HTREEITEM GetRootSel(HWND treeview)
{
	HTREEITEM selected, parent;

	selected = TreeView_GetNextItem(treeview, NULL, TVGN_CARET);
	parent = TreeView_GetNextItem(treeview, selected, TVGN_PARENT);

	return parent ? parent : selected;
}

#endif //WC_TREEVIEW

#ifdef TOOLTIPS_CLASS

/* Tooltips */
void TooltipInit(HWND tt);
BOOL AddTool(HWND dialog, int ctrlId, LPTSTR text);

/* Special Edit control subclass */
LRESULT CALLBACK TVEditWndProc(HWND control, UINT msg, WPARAM wParam, LPARAM lParam);

#endif //TOOLTIPS_CLASS
